import logging
import asyncio
import copy
import json 
from typing import Dict, List, Tuple

from .pipeline import knowledge_qa_pipeline
from guardrail.metrics.utils.keys import init_keys

from guardrail.firewall.output_detectors.base_detector import Detector

logging.basicConfig(level=logging.INFO)

class HallucinationKBDetector(Detector):
    def __init__(self, 
                 data_link="",
                 text="",
                 embedding_link="",
                 foundation_model="gpt-3.5-turbo", 
                 search_type="local",
                 category="kbqa"):
        init_keys()
        self.foundation_model = foundation_model
        if data_link != "":
            self.kb = data_link
        elif text != "":
            self.kb = self.self.convert_to_json(text)
        elif embedding_link !="":
            self.kb = embedding_link
        else: 
            self.kb = None
            print("Knowledge base not provided...")
        self.pipelines = {
                            "kbqa_local": knowledge_qa_pipeline(
                                foundation_model, 10, "local", data=self.kb if text != "" else None, data_link=self.kb if data_link != "" else None,
                                embedding_link=self.kb if embedding_link != "" else None
                            )
                        }
        self.search_type = search_type
        self.category=category
        self.inputs = [
            {
                "category": self.category,
                "search_type": self.search_type,
                "data":self.kb if text != "" else None, 
                "data_link":self.kb if data_link != "" else None,
                "embedding_link":self.kb if embedding_link != "" else None
            }
        ]
    
    def convert_to_json(self, input_text_or_path):
        if input_text_or_path.endswith('.json'):
            # Read the JSON file
            try:
                with open(input_text_or_path, 'r') as json_file:
                    data = json.load(json_file)
            except FileNotFoundError:
                return None
        else:
            # Treat the input as text and create a JSON structure
            data = {"text": input_text_or_path}

        return json.dumps(data, indent=4)
    
    def scan(self, prompt: str, output: str) -> Tuple[str, bool, float]:
        self.inputs[0]["prompt"] = prompt
        self.inputs[0]["response"] = output

        factuality_results = self.run(self.inputs)
        corrected_response = self.get_corrected_factual_response(factuality_results)
        risk_score = factuality_results["average_claim_level_factuality"]
        return corrected_response, True, risk_score
    
    def get_corrected_factual_response(self, factuality_results):
        corrected_response = ""
        for detail_info in factuality_results["detailed_information"]:
            response = detail_info["response"]
            for claim, claim_factuality in zip(detail_info["claims"], detail_info["claim_level_factuality"]):
                if not claim_factuality["factuality"]:
                    # Replace false claims with corrections in the response
                    claim_text = claim["claim"]
                    correction_text = claim_factuality["correction"]
                    response = response.replace(claim_text, correction_text)
            corrected_response = response
        return corrected_response
    
    def run(self, inputs):
        outputs = copy.deepcopy(inputs)
        batches = []
        current_category = inputs[0]['category']
        current_search_type = inputs[0].get('search_type', None)
        current_data_link = inputs[0].get('data_link', None)
        current_embedding_link = inputs[0].get('embedding_link', None)
        current_batch = []

        for input in inputs:
            if (input['category'] == current_category != 'kbqa') \
                or (input['category'] == current_category == 'kbqa' and input.get('search_type', None) == current_search_type == "online") \
                    or (input['category'] == current_category == 'kbqa' and input.get('search_type', None) == current_search_type == "local"\
                        and input.get('data_link', None)==current_data_link and input.get('embedding_link', None)==current_embedding_link):
                current_batch.append(input)
            else:
                batches.append(current_batch)
                current_batch = [input]
                current_category = input['category']
                current_search_type = input.get('search_type', None)
                current_data_link = input.get('data_link', None)
                current_embedding_link = input.get('embedding_link', None)

        batches.append(current_batch)  # append the last batch

        index = 0
        for batch in batches:
            if not batch: continue
            category = batch[0]['category']
            search_type = batch[0].get('search_type', None)
            if category == 'code':
                batch_results = asyncio.run(
                    self.pipelines[category].run_with_tool_api_call(
                        [sample['prompt'] for sample in batch],
                        [sample['response'] for sample in batch],
                        [sample['entry_point'] for sample in batch]
                    )
                )
            elif category == 'kbqa':
                if search_type is None or search_type == "online":
                    batch_results = asyncio.run(
                        self.pipelines[category+"_online"].run_with_tool_api_call(
                            [sample['prompt'] for sample in batch],
                            [sample['response'] for sample in batch],
                        )
                    )
                else:
                    batch_results = asyncio.run(
                        knowledge_qa_pipeline(
                            self.foundation_model,2,"local",batch[0].get("data_link"),batch[0].get("embedding_link")
                        ).run_with_tool_api_call(
                            [sample['prompt'] for sample in batch],
                            [sample['response'] for sample in batch],
                        )
                    )
            else:
                batch_results = asyncio.run(
                    self.pipelines[category].run_with_tool_api_call(
                        [sample['prompt'] for sample in batch],
                        [sample['response'] for sample in batch]
                    )
                )
            for result in batch_results:
                outputs[index].update(result)
                index += 1

        # calculate average response_level_factuality
        total_response_factuality = sum(output['response_level_factuality'] == True for output in outputs)
        avg_response_level_factuality = total_response_factuality / len(outputs)

        # calculate average claim_level_factuality
        num_claims = 0
        total_claim_factuality = 0
        for output in outputs:
            num_claims += len(output['claim_level_factuality'])
            if output['category'] == 'kbqa':
                total_claim_factuality += sum(claim['factuality'] == True for claim in output['claim_level_factuality'])
            elif output['category'] == 'code':
                total_claim_factuality += (output['claim_level_factuality'] == True)
            elif output['category'] == 'math':
                total_claim_factuality += sum(claim_factuality == True for claim_factuality in output['claim_level_factuality'])
            elif output['category'] == 'scientific':
                total_claim_factuality += sum(claim['factuality'] == True for claim in output['claim_level_factuality'])

        avg_claim_level_factuality = total_claim_factuality / num_claims

        return {"average_claim_level_factuality": avg_claim_level_factuality, "average_response_level_factuality": avg_response_level_factuality, "detailed_information": outputs}

    async def run_for_plugin(self, inputs):
        outputs = copy.deepcopy(inputs)

        batches = []
        current_category = inputs[0]['category']
        current_batch = []

        for input in inputs:
            if input['category'] == current_category:
                current_batch.append(input)
            else:
                batches.append(current_batch)
                current_batch = [input]
                current_category = input['category']

        batches.append(current_batch)  # append the last batch

        index = 0
        for batch in batches:
            category = batch[0]['category']
            if category == 'code':
                batch_results = await self.pipelines[category].run_with_tool_api_call(
                    [sample['prompt'] for sample in batch],
                    [sample['response'] for sample in batch],
                    [sample['entry_point'] for sample in batch],
                )
            else:
                batch_results = await self.pipelines[category].run_with_tool_api_call(
                    [sample['prompt'] for sample in batch],
                    [sample['response'] for sample in batch],
                )
            for result in batch_results:
                outputs[index].update(result)
                index += 1

        return outputs
    
   


"""
Guardrail ML evaluation and monitoring library
"""
from guardrail.firewall.firewall import Firewall
from guardrail.tracker.openai import OpenAI
from guardrail.tracker.langchain import GuardrailCallbackHandler


import unittest
from guardrail.guardchain import Chain, BufferMemory, BaseLanguageModel, ChatAgent

class TestGuardChainInit(unittest.TestCase):
    test_prompt="Write me a poem about AI"

    def test_guardchain_init(self):
        llm = BaseLanguageModel(model_name="EleutherAI/pythia-410m-deduped-v0")
        memory = BufferMemory()
        agent = ChatAgent.from_llm_and_tools(llm=llm)
        chain = Chain(agent=agent, memory=memory)
        output = chain.run(self.test_prompt)['message']
        print(output)
        self.assertGreater(len(output), 1)

unittest.main()



import logging
import os
import re
import json
from typing import List, Optional

from guardrail.firewall.input_detectors.base_detector import Detector

log = logging.getLogger(__name__)
stop_file_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "..",
    "patterns",
    "prompt_stop_substrings.json",
)
allowed_match_type = ["str", "word"]

class StopInputSubstrings(Detector):
    """
    A text detector that checks and filters out banned substrings from generated text prompt.

    This class provides the functionality to filter substrings in two different ways: as 'str' or as 'word'.
    - 'str' filters out substrings found anywhere in the text.
    - 'word' filters out substrings found as whole words.
    """

    def __init__(
        self,
        match_type: str = "str",
        case_sensitive: bool = False,
        substrings: Optional[List[str]] = None,
    ):
        """
        Initializes FilterSubstrings with a match type, case sensitivity option, and a list of substrings.

        Parameters:
            match_type (str): The type of substring matching. Can be either 'str' or 'word'. Default is 'str'.
            case_sensitive (bool): Determines whether the substring matching is case sensitive. Default is False.
            substrings (Optional[List[str]]): The list of substrings to be filtered out from the text. Default is None.

        Raises:
            ValueError: If match_type is not 'str' or 'word'.
        """

        if match_type not in allowed_match_type:
            raise ValueError(f"This match_type is not recognized. Allowed: {allowed_match_type}")
        
        # Load default substrings
        default_substrings = self.load_default_substrings()

        if substrings:
            log.info("Using user-provided substrings in addition to default patterns.")
            self._substrings = default_substrings + substrings
        else:
            log.warning("Using default patterns...")
            self._substrings = default_substrings
        
        self._match_type = match_type  # str or word
        self._case_sensitive = case_sensitive

    def load_default_substrings(self) -> List[str]:
        """
        Load default substrings from a JSON file.

        Returns:
            List[str]: The default list of substrings.
        """
        default_substrings = []

        try:
            with open(stop_file_path, "r") as stop_file:
                data = json.load(stop_file)
                for key in data:
                    default_substrings.extend(data[key])
            return default_substrings
        except (FileNotFoundError, json.JSONDecodeError):
            log.warning("Failed to load default substrings. Using an empty list.")
            return []

    def scan(self, prompt) -> (str, bool, float):
        filtered_prompt = prompt
        matched_substrings = []
        for s in self._substrings:
            if self._case_sensitive:
                s, filtered_prompt = s.lower(), filtered_prompt.lower()

            if self._match_type == "str":
                filtered_prompt = filtered_prompt.replace(s, "")
                if s in prompt:
                    matched_substrings.append(s)
            elif self._match_type == "word":
                pattern = r"\b" + re.escape(s) + r"\b"
                filtered_prompt = re.sub(pattern, "", filtered_prompt)
                if re.search(pattern, prompt):
                    matched_substrings.append(s)

        if matched_substrings:
            log.warning(f"Filtered out the following banned substrings: {matched_substrings}")
            return filtered_prompt, False, 1.0

        log.debug("No banned substrings found")
        return filtered_prompt, True, 0.0



import unittest
from guardrail.metrics import textstat


class TestTextStat(unittest.TestCase):
    test_data = (
        "Playing games has always been thought to be important to "
        "the development of well-balanced and creative children; "
        "however, what part, if any, they should play in the lives "
        "of adults has never been researched that deeply. I believe "
        "that playing games is every bit as important for adults "
        "as for children. Not only is taking time out to play games "
        "with our children and other adults valuable to building "
        "interpersonal relationships but is also a wonderful way "
        "to release built up tension."
    )

    def test_automated_readability_index(self):
        self.assertEqual(textstat.aggregate_reading_level(self.test_data), 15.5)


unittest.main()



import openai
from ..offline_evals import prompts
from ..utils.keys import init_openai_key, init_guardrail_key
from ..utils.textutils import TextUtilities

class OpenAIEvaluator():
    model_engine: str

    def __init__(
        self, *args, model_engine="gpt-3.5-turbo", **kwargs
    ):
        self_kwargs = dict()
        self_kwargs.update(**kwargs)
        self_kwargs['model_engine'] = model_engine
        self.model_engine = model_engine
        self.text_utils = TextUtilities()

        load_api_key("OPENAI_API_KEY")

    def run_online_moderation_metrics(self, prompt, output):
        moderation_not_hate = self.moderation_not_hate(output)
        moderation_not_hatethreatening = self.moderation_not_hatethreatening(output)

    def _create_chat_completion(self, *args, **kwargs):
        return openai.ChatCompletion.create(*args, **kwargs)

    def _moderation(self, text: str):
        return openai.Moderation.create(input=text)
    
    def _run_chat_completion(self, messages: list[dict[str, str]]) -> str:
        response = self._create_chat_completion(
            model=self.model_engine,
            temperature=0.0,
            messages=messages
        )
        score = response["choices"][0]["message"]["content"]
        return response, score

    def _evaluate(self, text: str, system_prompt: str) -> float:
        messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": text}
        ]
        response, score = self._run_chat_completion(messages)
        # openai_api_calculate_cost(response.usage, self.model_engine)

        return self.text_utils.re_1_10_score(score) / 10

    def evaluate_text(self, text: str, system_prompt: str) -> float:
        return self._evaluate(text, system_prompt)
    
    ### OpenAI Moderation
    def _calculate_inverted_score(self, category_name, text: str) -> float:
        openai_response = self._moderation(text)
        results = openai_response["results"]
        score = results[0]["category_scores"][category_name]
        return 1 - score, results

    def moderation_not_hate(self, text: str) -> float:
        return self._calculate_inverted_score("hate", text)

    def moderation_not_hatethreatening(self, text: str) -> float:
        return self._calculate_inverted_score("hate/threatening", text)

    def moderation_not_selfharm(self, text: str) -> float:
        return self._calculate_inverted_score("self-harm", text)

    def moderation_not_sexual(self, text: str) -> float:
        return self._calculate_inverted_score("sexual", text)

    def moderation_not_sexualminors(self, text: str) -> float:
        return self._calculate_inverted_score("sexual/minors", text)

    def moderation_not_violence(self, text: str) -> float:
        return self._calculate_inverted_score("violence", text)

    def moderation_not_violencegraphic(self, text: str) -> float:
        return self._calculate_inverted_score("violence/graphic", text)

    def moderation_not_harassment(self, text: str) -> float:
        return self._calculate_inverted_score("harassment", text)

    def moderation_not_harassmentthreatening(self, text: str) -> float:
        return self._calculate_inverted_score("harassment/threatening", text)

    ### Define LangChain evaluation methods
    def conciseness(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_CONCISENESS_PROMPT)

    def correctness(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_CORRECTNESS_PROMPT)

    def coherence(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_COHERENCE_PROMPT)

    def harmfulness(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_HARMFULNESS_PROMPT)

    def maliciousness(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_MALICIOUSNESS_PROMPT)

    def helpfulness(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_HELPFULNESS_PROMPT)

    def controversiality(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_CONTROVERSIALITY_PROMPT)

    def misogyny(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_MISOGYNY_PROMPT)

    def criminality(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_CRIMINALITY_PROMPT)

    def insensitivity(self, text: str) -> float:
        return self.evaluate_text(text, prompts.LANGCHAIN_INSENSITIVITY_PROMPT)


import setuptools
from setuptools import find_packages

import os

# Get the absolute path of the directory containing setup.py
current_dir = os.path.dirname(os.path.abspath(__file__))

# Construct the path to requirements.txt in the parent directory
parent_dir = os.path.abspath(os.path.join(current_dir, os.pardir))
requirements_path = "C:/Users/sophi/OneDrive/Documents/Data Science/guardrail/requirements.txt"

requirements = []
try:
    with open(requirements_path) as f:
        requirements = f.read().splitlines()
except FileNotFoundError:
    print(f"Error: 'requirements.txt' file not found at {requirements_path}.")
except Exception as e:
    print(f"An error occurred while reading 'requirements.txt': {e}")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="guardrail-ml",
    version="0.0.14",
    packages=find_packages(exclude=["tests", "guardrailmlenv", "examples", "docs", "env", "dist"]),
    package_data={'guardrail': ['firewall/output_detectors/factuality_detector/prompts/*.yaml', 'firewall/patterns/*.json']},
    author="Kevin Wu",
    author_email="kevin@guardrailml.com",
    description="Monitor LLMs with custom metrics to scale with confidence",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://www.guardrailml.com",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "PyPDF2",
        "textstat",
        "transformers<=4.30.2",
        "sentencepiece",
        "sentence-transformers",
        "accelerate",
        "bitsandbytes",
        "cleantext",
        "unidecode",
        "pillow",
        "jsonformer",
        "scipy",
        "pydantic==2.0.3",  # Specifying the version
        "tenacity",
        "colorama",
        "openai",
        "python-dotenv",
        "einops",
        "langchain",
        "Faker",
        "tensorflow<2.14.0",  # Specifying the version
        "presidio-analyzer",  # Specifying the version
        "presidio-anonymizer",  # Specifying the version
        "genbit",  # Specifying the version
        "jsonlines",
        "pydantic-settings<=2.0.3",  # Specifying the version
        "detect-secrets",
        "tiktoken",
        "spacy-transformers",
    ],
)



import enum
from typing import List, Any, Dict, Optional

from pydantic import BaseModel, Field


class MessageType(enum.Enum):
    UserMessage = enum.auto()
    AIMessage = enum.auto()
    SystemMessage = enum.auto()
    FunctionMessage = enum.auto()


class BaseMessage(BaseModel):
    """Message object."""

    content: str
    additional_kwargs: Optional[Dict[str, Any]] = None  # Use Optional with default None

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        raise NotImplementedError("Subclasses must implement this property.")


class UserMessage(BaseMessage):
    """Type of message that is spoken by the human."""

    content: str  # Ensure that content is defined with type str
    example: bool = False

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "user"


class AIMessage(BaseMessage):
    """Type of message that is spoken by the AI."""

    example: bool = False
    function_call: Dict[str, Any] = {}

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "ai"


class SystemMessage(BaseMessage):
    """Type of message that is a system message."""

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "system"


class FunctionMessage(BaseMessage):
    """Type of message that is a function message."""

    name: str

    @property
    def type(self) -> str:
        """Type of the message, used for serialization."""
        return "function"


class ChatMessageHistory(BaseModel):
    messages: List[BaseMessage] = []

    def save_message(self, message: str, message_type: MessageType, **kwargs):
        message_class = {
            MessageType.AIMessage: AIMessage,
            MessageType.UserMessage: UserMessage,
            MessageType.FunctionMessage: FunctionMessage,
            MessageType.SystemMessage: SystemMessage,
        }.get(message_type)

        if message_class:
            self.messages.append(message_class(content=message, **kwargs))

    def format_message(self) -> str:
        string_messages = []
        for m in self.messages:
            role = {
                UserMessage: "User",
                AIMessage: "Assistant",
                SystemMessage: "System",
            }.get(type(m))
            if role:
                string_messages.append(f"{role}: {m.content}")
        return "\n".join(string_messages) + "\n"

    def get_latest_user_message(self) -> UserMessage:
        for message in reversed(self.messages):
            if isinstance(message, UserMessage):
                return message
        return UserMessage(content="n/a")

    def clear(self) -> None:
        self.messages = []



import torch
from transformers import AutoTokenizer, AutoModelForSequenceClassification


class PromptInjection:
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained("JasperLS/gelectra-base-injection")
        self.model = AutoModelForSequenceClassification.from_pretrained(
            "JasperLS/gelectra-base-injection"
        )

    def evaluate(self, text):
        inputs = self.tokenizer(text, return_tensors="pt")
        outputs = self.model(**inputs)
        probabilities = torch.softmax(outputs.logits, dim=1)
        return probabilities[0, 1].item()



import logging
import inspect
import traceback
from typing import Dict, Union

from colorama import Fore
from guardrail.guardchain.agent.message import MessageType
from guardrail.guardchain.agent.agent_structs import AgentAction, AgentFinish
from guardrail.guardchain.chain.base_chain import BaseChain
from guardrail.guardchain.errors import ToolRunningError
from guardrail.guardchain.tools.base_tool import Tool
from guardrail.guardchain.tools.simple_handoff import HandOffToAgent

logger = logging.getLogger(__name__)


class Chain(BaseChain):
    """
    Default chain with take_next_step implemented.
    It handles a few common error cases with the agent, such as taking repeated action with the same
    inputs and whether the agent should continue the conversation.
    """

    return_intermediate_steps: bool = False
    handle_parsing_errors: bool = True  # Add type annotation 'bool'
    graceful_exit_tool: Tool = HandOffToAgent()

    def handle_repeated_action(self, agent_action: AgentAction) -> AgentFinish:
        print(f"Action taken before: {agent_action.tool}, " f"input: {agent_action.tool_input}")
        if agent_action.model_response:
            return AgentFinish(
                message=agent_action.response,
                log=f"Action taken before: {agent_action.tool}, "
                f"input: {agent_action.tool_input}",
            )
        else:
            print("No response from the agent. Gracefully exit due to repeated action")
            return AgentFinish(
                message=self.graceful_exit_tool.run(),
                log=f"Gracefully exit due to repeated action",
            )

    def take_next_step(
        self,
        name_to_tool_map: Dict[str, Tool],
        inputs: Dict[str, str],
    ) -> Union[AgentFinish, AgentAction]:
        """
        How the agent determines the next step after observing the inputs and intermediate steps.
        Args:
            name_to_tool_map: a map of tool name to the actual tool object.
            inputs: a dictionary of all inputs, such as user query, past conversation, and
                tool outputs.

        Returns:
            Either an AgentFinish to respond to the user or an AgentAction to take the next action.
        """

        try:
            # Call the LLM to see what to do.
            output = self.agent.plan(
                **inputs,
            )
        except Exception as e:
            if not self.handle_parsing_errors:
                raise e
            tool_output = f"Invalid or incomplete response due to {e}"
            traceback.print_exc()  # This will print the traceback with file name and line number.            

            current_frame = inspect.currentframe()
            line_number = current_frame.f_lineno
            file_name = inspect.getfile(current_frame)
            print(tool_output)
            print(f"File: {file_name}, Line: {line_number}, {tool_output}")
            output = AgentFinish(message=self.graceful_exit_tool.run(), log=tool_output)
            return output

        if isinstance(output, AgentAction):
            output = self.agent.clarify_args_for_agent_action(output, **inputs)

        # If the agent plans to respond with AgentFinish or there is a clarifying question, respond to
        # the user by returning AgentFinish.
        if isinstance(output, AgentFinish):
            return output

        if isinstance(output, AgentAction):
            tool_output = ""
            # Check if the tool is supported.
            if output.tool in name_to_tool_map:
                tool = name_to_tool_map[output.tool]

                # How to handle the case where the same action with the same input is taken before.
                if output.tool_input == self.memory.load_memory(tool.name):
                    return self.handle_repeated_action(output)

                self.memory.save_memory(tool.name, output.tool_input)
                # We then call the tool on the tool input to get a tool_output.
                try:
                    tool_output = tool.run(output.tool_input)
                except ToolRunningError as e:
                    new_agent_action = self.agent.fix_action_input(tool, output, error=str(e))
                    if new_agent_action and new_agent_action.tool_input != output.tool_input:
                        tool_output = tool.run(output.tool_input)

                print(
                    f"Took action '{tool.name}' with inputs '{output.tool_input}', "
                    f"and the tool_output is {tool_output}"
                )
            else:
                tool_output = f"Tool {output.tool} is not supported"

            output.tool_output = tool_output
            return output
        else:
            raise ValueError(f"Unsupported action: {type(output)}")



from sentence_transformers import SentenceTransformer, util

# Used for evaluating model on training / test datasets
class Similarity:
    def __init__(self, model_name="sentence-transformers/paraphrase-MiniLM-L6-v2"):
        self.embedder = SentenceTransformer(model_name)

    def evaluate(self, response, answer, correct_threshold=0.3):
        response_embedding = self.embedder.encode(response, convert_to_tensor=True)
        correct_answer_embedding = self.embedder.encode(answer, convert_to_tensor=True)
        
        cosine_similarity = util.pytorch_cos_sim(response_embedding, correct_answer_embedding).item()
        print('Cosine similarity is ' + str(cosine_similarity))

        # Adjust the threshold as needed, >30% threshold
        is_correct = cosine_similarity >= correct_threshold 

        return is_correct, cosine_similarity



"""OpenAI chat wrapper."""
from __future__ import annotations

import enum
import inspect
import logging
import os
import re
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

from pydantic import Extra, Field, root_validator

from guardchain.agent.message import (
    BaseMessage,
    UserMessage,
    AIMessage,
    SystemMessage,
    FunctionMessage,
)
from guardchain.models.base_model import (
    LLMResult,
    Generation,
    BaseLanguageModel,
)
from guardchain.tools.base_tool import Tool

logger = logging.getLogger(__name__)


def convert_dict_to_message(_dict: dict) -> BaseMessage:
    role = _dict["role"]
    if role == "user":
        return UserMessage(content=_dict["content"])
    elif role == "assistant":
        return AIMessage(
            content=_dict["content"] or "hum..",
            function_call=_dict.get("function_call", {}),
        )
    elif role == "system":
        return SystemMessage(content=_dict["content"])
    else:
        raise ValueError(f"Unsupported role {role}")


def convert_message_to_dict(message: BaseMessage) -> dict:
    if isinstance(message, UserMessage):
        message_dict = {"role": "user", "content": message.content}
    elif isinstance(message, AIMessage):
        message_dict = {"role": "assistant", "content": message.content}
    elif isinstance(message, SystemMessage):
        message_dict = {"role": "system", "content": message.content}
    elif isinstance(message, FunctionMessage):
        message_dict = {
            "role": "function",
            "content": message.content,
            "name": message.name,
        }
    else:
        raise ValueError(f"Got unknown type {message}")
    return message_dict


def convert_tool_to_dict(tool: Tool):
    """Convert tool into function parameter for openai"""
    inspection = inspect.getfullargspec(tool.func)
    arg_description = tool.arg_description or {}

    def _type_to_string(t: type) -> str:
        prog = re.compile(r"<class '(\w+)'>")
        cls = prog.findall(str(t))

        primary_type_map = {"str": "string"}

        if len(cls) > 0:
            cls_name = cls[0].split(".")[-1]
            return primary_type_map.get(cls_name, cls_name)

        if issubclass(t, enum.Enum):
            return "enum"

        return str(t)

    def _format_property(t: type, arg_desp: str):
        p = {"type": _type_to_string(t)}
        if arg_desp:
            p["description"] = arg_desp

        return p

    arg_annotations = inspection.annotations
    if arg_annotations:
        properties = {
            arg: _format_property(t, arg_description.get(arg))
            for arg, t in arg_annotations.items()
        }
    else:
        properties = {
            arg: _format_property(str, arg_description.get(arg))
            for arg in inspection.args
        }

    default_args = inspection.defaults or []
    required_args = inspection.args[: len(inspection.args) - len(default_args)]

    output = {
        "name": tool.name,
        "description": tool.description,
        "parameters": {
            "type": "object",
            "properties": properties,
            "required": required_args,
        },
    }

    return output


class OpenAIModel(BaseLanguageModel):
    """Wrapper around OpenAI Chat large language models.

    To use, you should have the ``openai`` python package installed, and the
    environment variable ``OPENAI_API_KEY`` set with your API key.

    Any parameters that are valid to be passed to the openai.create call can be passed
    in, even if not explicitly saved on this class.

    Example:
        .. code-block:: python

            from guardchain.models.chat_openai import ChatOpenAI
            openai = ChatOpenAI()
    """

    client: Any  #: :meta private:
    model_name: str = "gpt-3.5-turbo"
    """Model name to use."""
    temperature: float = 0
    """What sampling temperature to use."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Holds any model parameters valid for `create` call not explicitly specified."""
    openai_api_key: Optional[str] = None
    openai_organization: Optional[str] = None
    api_type: Optional[str] = None
    """OpenAI API type, it can be `openai` or `azure`."""
    api_base: Optional[str] = None
    """The OpenAI API base url or Azure OpenAI API base url."""
    azure_api_version: Optional[str] = None
    """Azure API version."""
    azure_deployment_name: Optional[str] = None
    """Azure deployment name."""
    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    """Timeout for requests to OpenAI completion API. Default is 600 seconds."""
    max_retries: int = 6
    """Maximum number of retries to make when generating."""
    # TODO: support streaming
    # streaming: bool = False
    # """Whether to stream the results or not."""
    # n: int = 1
    # """Number of chat completions to generate for each prompt."""
    max_tokens: Optional[int] = None
    """Maximum number of tokens to generate."""

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.ignore

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that api key and python package exists in environment."""
        openai_api_key = os.environ["OPENAI_API_KEY"]
        openai_api_type = os.environ.get("OPENAI_API_TYPE", "open_ai")
        openai_api_base = os.environ.get("OPENAI_API_BASE", None)
        try:
            import openai

        except ImportError:
            raise ValueError(
                "Could not import openai python package. "
                "Please install it with `pip install openai`."
            )
        values["api_key"] = openai.api_key = openai_api_key
        values["api_type"] = openai.api_type = openai_api_type
        if openai_api_base:
            values["api_base"] = openai.api_base = openai_api_base
        if openai_api_type == "azure":
            values["azure_api_version"] = openai.api_version = os.environ.get("OPENAI_API_VERSION", "2023-05-15")
        try:
            values["client"] = openai.ChatCompletion
        except AttributeError:
            raise ValueError(
                "`openai` has no `ChatCompletion` attribute, this is likely "
                "due to an old version of the openai package. Try upgrading it "
                "with `pip install --upgrade openai`."
            )
        # if values["n"] < 1:
        #     raise ValueError("n must be at least 1.")
        # if values["n"] > 1 and values["streaming"]:
        #     raise ValueError("n must be 1 when streaming.")
        return values

    def generate(
        self,
        messages: List[BaseMessage],
        functions: Optional[List[Tool]] = None,
        stop: Optional[List[str]] = None,
    ) -> LLMResult:
        message_dicts, function_dicts, params = self._create_message_dicts(
            messages, functions, stop
        )

        generation_param = {
            "messages": message_dicts,
            **params,
        }
        if len(function_dicts) > 0:
            generation_param["functions"] = function_dicts

        response = self.generate_with_retry(**generation_param)
        return self._create_llm_result(response)

    def _create_message_dicts(
        self,
        messages: List[BaseMessage],
        tools: Optional[List[Tool]],
        stop: Optional[List[str]],
    ) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]], Dict[str, Any]]:
        params: Dict[str, Any] = {**{"model": self.model_name}, **self._default_params}
        if self.azure_deployment_name and self.api_type == "azure":
            params["engine"] = self.azure_deployment_name
        if stop is not None:
            if "stop" in params:
                raise ValueError("`stop` found in both the input and default params.")
            params["stop"] = stop
        message_dicts = [convert_message_to_dict(m) for m in messages]
        function_dicts = []
        if tools:
            function_dicts = [convert_tool_to_dict(t) for t in tools]
        return message_dicts, function_dicts, params

    def _create_llm_result(self, response: Mapping[str, Any]) -> LLMResult:
        generations = []
        for res in response["choices"]:
            message = convert_dict_to_message(res["message"])
            gen = Generation(message=message)
            generations.append(gen)
        llm_output = {"token_usage": response["usage"], "model_name": self.model_name}
        result = LLMResult(generations=generations, llm_output=llm_output)
        return result


import yaml
from .openai_wrapper import OpenAIChat
import os
import pathlib

class pipeline():
    def __init__(self, domain, llm_model="gpt-3.5-turbo"):
        #if foundation_model == 'gpt-3.5-turbo' or foundation_model == 'gpt-4':
        self.company = 'openai'
        self.chat = OpenAIChat(model_name=llm_model)

        self.prompts_path = os.path.join(os.path.dirname(pathlib.Path(__file__)), "../prompts/")
        
        with open(os.path.join(self.prompts_path, "self_check.yaml"), 'r') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        self.self_check_prompt = data[domain]


from .chain.chain import Chain
from .memory.buffer_memory import BufferMemory
from .models.base_model import BaseLanguageModel
from .models.ggml_model import GGMLHuggingFaceModel
from .models.openai_model import ChatOpenAI
from .agent.chat_agent import ChatAgent



import logging
import time
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

from colorama import Fore
from guardrail.guardchain.agent.base_agent import BaseAgent
from guardrail.guardchain.agent.message import ChatMessageHistory, MessageType
from guardrail.guardchain.agent.agent_structs import AgentAction, AgentFinish
from guardrail.guardchain.chain import constants
from guardrail.guardchain.memory.base import BaseMemory
from guardrail.guardchain.tools.base_tool import Tool
from pydantic import BaseModel

logger = logging.getLogger(__name__)

class BaseChain(BaseModel, ABC):
    """
    Base interface that all chains should implement.
    Chain will standardize inputs and outputs, the main entry point is the run function.
    """

    agent: Optional[BaseAgent] = None
    memory: Optional[BaseMemory] = None
    last_query: str = ""
    max_iterations: Optional[int] = 15
    max_execution_time: Optional[float] = None

    def prep_inputs(self, user_query: str) -> Dict[str, str]:
        """Load conversation history from memory and prepare inputs."""
        inputs = {
            constants.CONVERSATION_HISTORY: ChatMessageHistory(),
            constants.INTERMEDIATE_STEPS: [],
        }
        if self.memory is not None:
            intermediate_steps = self.memory.load_memory(constants.INTERMEDIATE_STEPS, [])
            self.memory.save_conversation(message=user_query, message_type=MessageType.UserMessage)

            inputs[constants.CONVERSATION_HISTORY] = deepcopy(self.memory.load_conversation())
            inputs[constants.INTERMEDIATE_STEPS] = deepcopy(intermediate_steps)

        return inputs

    def prep_output(
        self,
        inputs: Dict[str, str],
        output: AgentFinish,
        return_only_outputs: bool = False,
    ) -> Dict[str, Any]:
        """Save conversation into memory and prepare outputs."""
        output_dict = output.format_output()
        if self.memory is not None:
            self.memory.save_conversation(
                message=output.message, message_type=MessageType.AIMessage
            )
            self.memory.save_memory(
                key=constants.INTERMEDIATE_STEPS, value=output.intermediate_steps
            )

        if return_only_outputs:
            return output_dict
        else:
            return {**inputs, **output_dict}

    def run(
        self,
        user_query: str,
        return_only_outputs: bool = False,
    ) -> Dict[str, Any]:
        """Wrapper for _run function by formatting the input and outputs.

        Args:
            user_query: user query
            return_only_outputs: boolean for whether to return only outputs in the
                response. If True, only new keys generated by this chain will be
                returned. If False, both input keys and new keys generated by this
                chain will be returned. Defaults to False.

        """
        inputs = self.prep_inputs(user_query)
        logger.info(f"\nInput to agent: {inputs}")
        try:
            output = self._run(inputs)
        except (KeyboardInterrupt, Exception) as e:
            raise e

        return self.prep_output(inputs, output, return_only_outputs)

    def _run(
        self,
        inputs: Dict[str, Any],
    ) -> AgentFinish:
        """
        Run inputs including user query and past conversation with agent and get response back.
        Calls take_next_step function to determine what should be the next step after
        collecting all the inputs and memorized contents.
        """
        # Construct a mapping of tool name to tool for easy lookup
        name_to_tool_map = {tool.name: tool for tool in self.agent.tools}

        intermediate_steps: List[AgentAction] = inputs[constants.INTERMEDIATE_STEPS]

        # Let's start tracking the number of iterations and time elapsed
        iterations = 0
        time_elapsed = 0.0
        start_time = time.time()
        # We now enter the agent loop (until it returns something).
        while self._should_continue(iterations, time_elapsed):
            logger.info(f"\nIntermediate steps: {intermediate_steps}\n")
            next_step_output = self.should_answer(inputs=inputs)

            # if next_step_output is None, which means the agent should answer and take next step
            if not next_step_output:
                next_step_output = self.take_next_step(
                    name_to_tool_map,
                    inputs,
                )

            if isinstance(next_step_output, AgentFinish):
                next_step_output.intermediate_steps = intermediate_steps
                return next_step_output

            # stores action output into the conversation as FunctionMessage, which can be used by
            # OpenAIFunctionsAgent
            if isinstance(next_step_output, AgentAction):
                self.memory.save_conversation(
                    message=str(next_step_output.tool_output),
                    name=next_step_output.tool,
                    message_type=MessageType.FunctionMessage,
                )

            intermediate_steps.append(next_step_output)
            # update inputs
            inputs[constants.INTERMEDIATE_STEPS] = intermediate_steps
            inputs[constants.CONVERSATION_HISTORY] = self.memory.load_conversation()

            iterations += 1
            time_elapsed = time.time() - start_time

        # force termination when shouldn't continue
        output = AgentFinish(
            message="Agent stopped due to iteration limit or time limit.",
            log="",
            intermediate_steps=intermediate_steps,
        )
        return output

    @abstractmethod
    def take_next_step(
        self,
        name_to_tool_map: Dict[str, Tool],
        inputs: Dict[str, str],
    ) -> Union[AgentFinish, AgentAction]:
        """How agent determines the next step after observing the inputs and intermediate
        steps"""

    def _should_continue(self, iterations: int, time_elapsed: float) -> bool:
        if self.max_iterations is not None and iterations >= self.max_iterations:
            return False
        if self.max_execution_time is not None and time_elapsed >= self.max_execution_time:
            return False

        return True

    def should_answer(self, inputs) -> Optional[AgentFinish]:
        """
        Let the agent determine if it should continue to answer questions
        or if the conversation should end.

        Args:
            inputs: Dict containing user query and other memorized contents

        Returns:
            None if the agent should answer
            AgentFinish if the agent should NOT answer and respond to the user with a message
        """
        output = None
        # check if agent should answer this query
        last_query = inputs[constants.CONVERSATION_HISTORY].get_latest_user_message().content
        if self.last_query != last_query:
            output = self.agent.should_answer(**inputs)
            self.last_query = last_query

        return output



import numpy as np
from guardrail.dataset.embeddings import EmbeddingModel

# Test case 1: Testing the `generate_freshness_score` method
embedding_data = {
    'result1': np.array([0.8, 0.2, 0.5]),
    'result2': np.array([0.9, 0.3, 0.6]),
    'result3': np.array([0.7, 0.1, 0.4]),
    # Add more text data and corresponding embeddings here
}

embedding_model = EmbeddingModel(embedding_data, non_linear_function='sigmoid')
query_embedding = np.array([0.8, 0.2, 0.5])
freshness_score = embedding_model.generate_freshness_score(query_embedding, weight=1.5, bias=-0.5)
print("Freshness Score:", freshness_score)

# Test case 2: Testing the `improve_embeddings_with_diversity` method
query_embedding = np.array([0.8, 0.2, 0.5])
diverse_results = embedding_model.improve_embeddings_with_diversity(query_embedding, diversity_weight=0.7, top_k=3)
print("Diverse Results:", diverse_results)

# Test case 3: Testing the `reduce_embedding_dimension_pca` method
embeddings_data = np.random.rand(100, 50)  # Example embeddings matrix with 100 samples and 50 dimensions
target_dim_pca = 10  # The desired dimensionality of the reduced embeddings (PCA)
reduced_embeddings_pca = EmbeddingModel.reduce_embedding_dimension_pca(embeddings_data, target_dim_pca)
print("Reduced Embeddings Shape (PCA):", reduced_embeddings_pca.shape)

# Test case 4: Testing the `reduce_embedding_dimension_tsne` method
target_dim_tsne = 2  # The desired dimensionality of the reduced embeddings (t-SNE)
reduced_embeddings_tsne = EmbeddingModel.reduce_embedding_dimension_tsne(embeddings_data, target_dim_tsne)
print("Reduced Embeddings Shape (t-SNE):", reduced_embeddings_tsne.shape)

# Test case 5: Testing the `reduce_embedding_dimension_umap` method
target_dim_umap = 3  # The desired dimensionality of the reduced embeddings (UMAP)
reduced_embeddings_umap = EmbeddingModel.reduce_embedding_dimension_umap(embeddings_data, target_dim_umap)
print("Reduced Embeddings Shape (UMAP):", reduced_embeddings_umap.shape)

import numpy as np

# Test case 1: Diversity
embedding_data_diversity = {
    'get python': np.array([0.8, 0.2, 0.5]),
    'HTTP method': np.array([0.9, 0.3, 0.6]),
    'places to purchase a live python': np.array([0.7, 0.1, 0.4]),
    # Add more text data and corresponding embeddings here
}

embedding_model_diversity = EmbeddingModel(embedding_data_diversity, non_linear_function='sigmoid')
query_embedding_diversity = np.array([0.8, 0.2, 0.5])

# Set diversity_weight to 0.7 to prioritize diversity over relevance
diverse_results = embedding_model_diversity.improve_embeddings_with_diversity(query_embedding_diversity, diversity_weight=0.7, top_k=3)
print("Diverse Results:", diverse_results)
# Expected output: The diverse_results should contain a mix of relevant and diverse results.

# Test case 2: Non-linear scoring (Freshness)
embedding_data_freshness = {
    'election results today': np.array([0.9, 0.2, 0.6]),
    'election results yesterday': np.array([0.8, 0.1, 0.4]),
    'election results seven days ago': np.array([0.7, 0.05, 0.3]),
    # Add more text data and corresponding embeddings here
}

embedding_model_freshness = EmbeddingModel(embedding_data_freshness, non_linear_function='sigmoid')

# Test for today's data
query_embedding_freshness_today = np.array([0.9, 0.2, 0.6])
freshness_score_today = embedding_model_freshness.generate_freshness_score(query_embedding_freshness_today, weight=1.5, bias=-0.5)
print("Freshness Score (Today's data):", freshness_score_today)
# Expected output: The freshness_score_today should be high due to the non-linear scoring for today's data.

# Test for data from yesterday
query_embedding_freshness_yesterday = np.array([0.8, 0.1, 0.4])
freshness_score_yesterday = embedding_model_freshness.generate_freshness_score(query_embedding_freshness_yesterday, weight=1.5, bias=-0.5)
print("Freshness Score (Yesterday's data):", freshness_score_yesterday)
# Expected output: The freshness_score_yesterday should be lower than today's score due to non-linear scoring.

# Test for data from seven days ago
query_embedding_freshness_seven_days_ago = np.array([0.7, 0.05, 0.3])
freshness_score_seven_days_ago = embedding_model_freshness.generate_freshness_score(query_embedding_freshness_seven_days_ago, weight=1.5, bias=-0.5)
print("Freshness Score (Seven days ago data):", freshness_score_seven_days_ago)
# Expected output: The freshness_score_seven_days_ago should be similar to yesterday's score, indicating similar value in the non-linear scoring.

# Test case 3: Invalid non-linear function
try:
    embedding_model_invalid_function = EmbeddingModel(embedding_data_freshness, non_linear_function='invalid_function')
except ValueError as e:
    print("Error:", str(e))
# Expected output: An error message indicating that the non-linear function provided is invalid.




import json
import yaml
import os
import time
import math
import pdb
from typing import List, Dict

from .utils import pipeline
from .tools import google_search, local_search

class knowledge_qa_pipeline(pipeline):
    def __init__(self, foundation_model, snippet_cnt, search_type, data_link=None, embedding_link=None, data=None):
        super().__init__('knowledge_qa', foundation_model)
        self.foundation_model = foundation_model

        if(search_type == 'online'):
            self.tool = google_search(snippet_cnt = snippet_cnt)
        elif(search_type == 'local'):
            self.tool = local_search(snippet_cnt = snippet_cnt, data=data, data_link=data_link, embedding_link=embedding_link)
        with open(os.path.join(self.prompts_path, "claim_extraction.yaml"), 'r') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        self.claim_prompt = data['knowledge_qa']

        with open(os.path.join(self.prompts_path, 'query_generation.yaml'), 'r') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        self.query_prompt = data['knowledge_qa']

        with open(os.path.join(self.prompts_path, 'agreement_verification.yaml'), 'r') as file:
            data = yaml.load(file, Loader=yaml.FullLoader)
        self.verification_prompt = data['knowledge_qa']

    async def _claim_extraction(self, responses):
        messages_list = [
            [
                {"role": "system", "content": self.claim_prompt['system']},
                {"role": "user", "content": self.claim_prompt['user'].format(input=response)},
            ]
            for response in responses
        ]
        return await self.chat.async_run(messages_list, List)

    async def _query_generation(self, claims):
        if claims == None:
            return ['None']
        messages_list = [
            [
                {"role": "system", "content": self.query_prompt['system']},
                {"role": "user", "content": self.query_prompt['user'].format(input=claim['claim'] if 'claim' in claim else '')},
            ]
            for claim in claims
        ]
        return await self.chat.async_run(messages_list, List)

    async def _verification(self, claims, evidences):
        messages_list = [
            [
                {"role": "system", "content": self.verification_prompt['system']},
                {"role": "user", "content": self.verification_prompt['user'].format(claim=claim['claim'], evidence=str(evidence))},
            ]
            for claim, evidence in zip(claims, evidences)
        ]
        return await self.chat.async_run(messages_list, dict)

    async def run_with_tool_live(self, responses):
        claims_in_responses = await self._claim_extraction(responses)
        queries_in_responses = []
        evidences_in_responses = []
        sources_in_responses = []
        verifications_in_responses = []
        for claims_in_response in claims_in_responses:
            queries = await self._query_generation(claims_in_response)
            queries_in_responses.append(queries)
            search_outputs_for_claims = await self.tool.run(queries)
            print("Search claims: ", search_outputs_for_claims)
            evidences = [[output['content'] for output in search_outputs_for_claim] for search_outputs_for_claim in search_outputs_for_claims]
            evidences_in_responses.append(evidences)
            sources = [[output['source'] for output in search_outputs_for_claim] for search_outputs_for_claim in search_outputs_for_claims]
            sources_in_responses.append(sources)
            verifications = await self._verification(claims_in_response, evidences)
            verifications_in_responses.append(verifications)

        return claims_in_responses, queries_in_responses, evidences_in_responses, sources_in_responses, verifications_in_responses

    async def run_with_tool_live_without_claim_extraction(self, claims):
        queries = await self._query_generation(claims)
        evidences = await self.tool.run(queries)

        final_response = await self._verification(claims, evidences)
        for i in range(len(final_response)):
            if final_response[i] != None:
                final_response[i]['queries'] = queries[i]
                final_response[i]['evidences'] = evidences[i]

        return final_response

    async def run_with_tool_api_call(self, prompts, responses):
        batch_size = 5
        num_batches = math.ceil(len(prompts) / batch_size)

        self.sample_list = [{"prompt": prompt, "response": response, "category": 'kbqa'} for prompt, response in zip(prompts, responses)]

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = min((i + 1) * batch_size, len(responses))

            claims_in_responses, queries_in_responses, evidences_in_responses, sources_in_responses, verifications_in_responses = await self.run_with_tool_live(responses[batch_start:batch_end])

            for j, (claims_in_response, queries_in_response, evidences_in_response, sources_in_response, verifications_in_response) in enumerate(zip(claims_in_responses, queries_in_responses, evidences_in_responses, sources_in_responses, verifications_in_responses)):
                index = batch_start + j

                if claims_in_response != None:
                    for k, claim in enumerate(claims_in_response):
                        if verifications_in_response[k] != None:
                            if claim != None:
                                verifications_in_response[k].update({'claim': claim['claim']})
                            else:
                                verifications_in_response[k].update({'claim': 'None'})

                evidences_with_source = []
                for evidence, source in zip(evidences_in_response, sources_in_response):
                    evidences_with_source.append({'evidence': evidence, 'source': source})
                self.sample_list[index].update({
                    'claims': claims_in_response,
                    'queries': queries_in_response,
                    # 'evidences': evidences_in_response,
                    # 'sources': sources_in_response,
                    'evidences': evidences_with_source,
                    'claim_level_factuality': verifications_in_response,
                    'response_level_factuality': all([verification['factuality'] if verification != None else True for verification in verifications_in_response])
                })

        return self.sample_list

    async def run_with_tool_dataset(self, annotated_dataset_path: str, with_tool_classified_dataset_path: str, rerun: bool = False, rerun_indices: list = []):
        data_path = with_tool_classified_dataset_path if rerun else annotated_dataset_path
        with open(data_path, 'r') as f:
            data = [json.loads(line) for line in f]
        self.sample_list = data if rerun else [claim for sample in data for claim in sample['claims']]
        rerun_elements = self.sample_list if not rerun else [self.sample_list[i] for i in rerun_indices]

        batch_size = 4
        num_batches = math.ceil(len(rerun_elements) / batch_size) # 5

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = min((i + 1) * batch_size, len(rerun_elements))

            responses = await self.run_with_tool_live_without_claim_extraction(rerun_elements[batch_start:batch_end])

            for j, response in enumerate(responses):
                index = batch_start + j if rerun == False else rerun_indices[batch_start + j]
                if response is None:
                    self.sample_list[index].update({
                        'with_tool_classification': 'None',
                        'with_tool_reasoning': 'None',
                        'queries': 'None',
                        'evidences': 'None'
                    })
                else:
                    self.sample_list[index].update({
                        'with_tool_classification': response.get('factuality', 'None'),
                        'with_tool_reasoning': response.get('reasoning', 'None'),
                        'queries': response.get('queries', 'None'),
                        'evidences': response.get('evidences', 'None')
                    })

            # save everything after each batch to prevent data loss
            with open(with_tool_classified_dataset_path, 'w') as f:
                for item in self.sample_list:
                    json_str = json.dumps(item)
                    f.write(json_str + '\n')

    async def run_self_check_live(self, fewshot, batch):
        user_prompt_key = 'user_3_shot_CoT' if fewshot else 'user_zero_shot_CoT'
        messages_list = [
            [
                {"role": "system", "content": self.self_check_prompt['system']},
                {"role": "user", "content": self.self_check_prompt[user_prompt_key].format(claim=response['claim'])},
            ]
            for response in batch
        ]
        return await self.chat.async_run(messages_list, Dict)

    async def run_self_check_dataset(self, annotated_dataset_path: str, self_check_classified_dataset_path: str, fewshot: bool = False, rerun: bool = False, rerun_indices: list = []):
        data_path = annotated_dataset_path if not rerun else self_check_classified_dataset_path
        with open(data_path, 'r') as f:
            data = [json.loads(line) for line in f]
        self.sample_list = data if rerun else [claim for sample in data for claim in sample['claims']]
        rerun_elements = self.sample_list if not rerun else [self.sample_list[i] for i in rerun_indices]

        batch_size = 10
        num_batches = math.ceil(len(rerun_elements) / batch_size)

        for i in range(num_batches):
            print(i)
            batch_start = i * batch_size
            batch_end = min((i + 1) * batch_size, len(rerun_elements))
            batch = rerun_elements[batch_start:batch_end]

            responses = await self.run_self_check_live(fewshot, batch)
            for j, response in enumerate(responses):
                index = batch_start + j if not rerun else rerun_indices[batch_start + j]
                if response is None:
                    self.sample_list[index].update({
                        'self_check_classification': 'None',
                        'self_check_reasoning': 'None'
                    })
                else:
                    self.sample_list[index].update({
                        'self_check_classification': response.get('factuality', 'None'),
                        'self_check_reasoning': response.get('reasoning', 'None')
                    })

            # save everything after each batch to prevent data loss
            with open(self_check_classified_dataset_path, 'w') as f:
                for item in self.sample_list:
                    json_str = json.dumps(item)
                    f.write(json_str + '\n')


from __future__ import annotations

import os
import yaml
import ast
import pdb
import asyncio
from typing import Any, List
import os
import pathlib
import openai
import re
import asyncio
from guardrail.metrics.utils.keys import init_keys
from guardrail.tracker.openai import OpenAI

class OpenAIChat():
    def __init__(
            self,
            model_name='gpt-4-8k',
            max_tokens=2500,
            temperature=0,
            top_p=1,
            request_timeout=120,
    ):
        if 'gpt' not in model_name:
            openai.api_base = "http://localhost:8000/v1"
        else:
            #openai.api_base = "https://api.openai.com/v1"
            openai.api_key = os.environ.get("OPENAI_API_KEY", None)
            assert openai.api_key is not None, "Please set the OPENAI_API_KEY environment variable."
            assert openai.api_key !='', "Please set the OPENAI_API_KEY environment variable."

        self.config = {
            'model_name': model_name,
            'max_tokens': max_tokens,
            'temperature': temperature,
            'top_p': top_p,
            'request_timeout': request_timeout,
        }
        self.openai_gr = OpenAI(firewall=None)

    def extract_list_from_string(self, input_string):
        # pattern = r'\[.*\]'  
        # result = re.search(pattern, input_string)
        # if result:
        #     return result.group()
        # else:
        #     return None
        start_index = input_string.find('[')  
        end_index = input_string.rfind(']') 

        if start_index != -1 and end_index != -1 and start_index < end_index:
            return input_string[start_index:end_index + 1]
        else:
            return None
        
    def extract_dict_from_string(self, input_string):
        start_index = input_string.find('{')
        end_index = input_string.rfind('}')

        if start_index != -1 and end_index != -1 and start_index < end_index:
            return input_string[start_index:end_index + 1]
        else:
            return None
    
    def _boolean_fix(self, output):
        return output.replace("true", "True").replace("false", "False")

    def _type_check(self, output, expected_type):
        try:
            output_eval = ast.literal_eval(output)
            if not isinstance(output_eval, expected_type):
                return None
            return output_eval
        except:
            '''
            if(expected_type == List):
                valid_output = self.extract_list_from_string(output)
                output_eval = ast.literal_eval(valid_output)
                if not isinstance(output_eval, expected_type):
                    return None
                return output_eval
            elif(expected_type == dict):
                valid_output = self.extract_dict_from_string(output)
                output_eval = ast.literal_eval(valid_output)
                if not isinstance(output_eval, expected_type):
                    return None
                return output_eval
            '''
            return None

    async def dispatch_openai_requests(
        self,
        messages_list,
    ) -> list[str]:
        """Dispatches requests to OpenAI API asynchronously.
        
        Args:
            messages_list: List of messages to be sent to OpenAI ChatCompletion API.
        Returns:
            List of responses from OpenAI API.
        """
        async def _request_with_retry(messages, retry=3):
            for _ in range(retry):
                try:
                    response = self.openai_gr.run_chat_completion(
                        model=self.config['model_name'],
                        messages=messages,
                        max_tokens=self.config['max_tokens'],
                        temperature=self.config['temperature'],
                        top_p=self.config['top_p'],
                        request_timeout=self.config['request_timeout'],
                    )
                    return response
                except openai.error.RateLimitError:
                    print('Rate limit error, waiting for 40 second...')
                    await asyncio.sleep(40)
                except openai.error.APIError:
                    print('API error, waiting for 1 second...')
                    await asyncio.sleep(1)
                except openai.error.Timeout:
                    print('Timeout error, waiting for 1 second...')
                    await asyncio.sleep(1)
                except openai.error.ServiceUnavailableError:
                    print('Service unavailable error, waiting for 3 second...')
                    await asyncio.sleep(3)
                except openai.error.APIConnectionError:
                    print('API Connection error, waiting for 3 second...')
                    await asyncio.sleep(3)

            return None

        async_responses = [
            _request_with_retry(messages)
            for messages in messages_list
        ]

        return await asyncio.gather(*async_responses)
    
    async def async_run(self, messages_list, expected_type):
        retry = 1
        responses = [None for _ in range(len(messages_list))]
        messages_list_cur_index = [i for i in range(len(messages_list))]

        while retry > 0 and len(messages_list_cur_index) > 0:
            print(f'{retry} retry left...')
            messages_list_cur = [messages_list[i] for i in messages_list_cur_index]
            
            predictions = await self.dispatch_openai_requests(
                messages_list=messages_list_cur,
            )

            preds = [self._type_check(self._boolean_fix(prediction['choices'][0]['message']['content']), expected_type) if prediction is not None else None for prediction in predictions]

            finised_index = []
            for i, pred in enumerate(preds):
                if pred is not None:
                    responses[messages_list_cur_index[i]] = pred
                    finised_index.append(messages_list_cur_index[i])
            
            messages_list_cur_index = [i for i in messages_list_cur_index if i not in finised_index]
            
            retry -= 1
        
        return responses

class OpenAIEmbedWrapper():
    def __init__(self):
        init_keys()
        assert openai.api_key is not None, "Please set the OPENAI_API_KEY environment variable."
        assert openai.api_key != '', "Please set the OPENAI_API_KEY environment variable."

    async def create_embedding(self, text, retry=3):
        for _ in range(retry):
            try:
                response = await openai.Embedding.acreate(input=text, model="text-embedding-ada-002")
                return response
            except openai.error.RateLimitError:
                print('Rate limit error, waiting for 1 second...')
                await asyncio.sleep(1)
            except openai.error.APIError:
                print('API error, waiting for 1 second...')
                await asyncio.sleep(1)
            except openai.error.Timeout:
                print('Timeout error, waiting for 1 second...')
                await asyncio.sleep(1)
        return None

    async def process_batch(self, batch, retry=3):
        tasks = [self.create_embedding(text, retry=retry) for text in batch]
        return await asyncio.gather(*tasks)